# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import icalendar
import datetime
import pytz
from dateutil.rrule import rrulestr, rruleset, rrule, DAILY
import dateutil.parser
import sys

from collections import defaultdict
from icalendar.prop import vDDDTypes

if sys.version_info[0] == 2:
    _EPOCH = datetime.datetime.utcfromtimestamp(0)
    _EPOCH_TZINFO = pytz.UTC.localize(_EPOCH)
    def timestamp(dt):
        """Return the time stamp of a datetime"""
        # from https://stackoverflow.com/a/35337826
        if dt.tzinfo:
            return (dt - _EPOCH_TZINFO).total_seconds()
        return (dt - _EPOCH).total_seconds()
else:
    def timestamp(dt):
        """Return the time stamp of a datetime"""
        return dt.timestamp() 

def is_event(component):
    """Return whether a component is a calendar event."""
    return isinstance(component, icalendar.cal.Event)

def convert_to_date(date):
    """converts a date or datetime to a date"""
    return datetime.date(date.year, date.month, date.day)

def convert_to_datetime(date, tzinfo):
    """converts a date to a datetime"""
    if isinstance(date, datetime.datetime):
        if date.tzinfo is None:
            if tzinfo is not None:
                return tzinfo.localize(date)
        else:
            assert tzinfo is not None, "Error in algorithm: tzinfo expected to be known."
        return date
    elif isinstance(date, datetime.date):
        return datetime.datetime(date.year, date.month, date.day, tzinfo=tzinfo)
    return date

def time_span_contains_event(span_start, span_stop, event_start, event_stop,
        include_start=True, include_stop=True, comparable=False):
    """Return whether an event should is included within a time span.

    - span_start and span_stop define the time span
    - event_start and event_stop define the event time
    - include_start defines whether events overlapping the start of the
        time span should be included
    - include_stop defines whether events overlapping the stop of the
        time span should be included
    - comparable indicates whether the dates can be compared.
        You can set it to True if you are sure you have timezones and
        date/datetime correctly or used make_comparable() before.
    """
    if not comparable:
        span_start, span_stop, event_start, event_stop = make_comparable((
            span_start, span_stop, event_start, event_stop
        ))
    assert event_start <= event_stop, "the event must start before it ends"
    assert span_start <= span_stop, "the time span must start before it ends"
    return (include_start or span_start <= event_start) and \
        (include_stop or event_stop <= span_stop) and \
        event_start <= span_stop and span_start <= event_stop \
        and (event_stop != span_start or event_start == span_start) \
        and (span_stop != event_start or span_stop == event_stop)


def make_comparable(dates):
    """Make an list or tuple of dates comparable.

    Returns an list.
    """
    tzinfo = None
    for date in dates:
        tzinfo = getattr(date, "tzinfo", None)
        if tzinfo is not None:
            break
    return [convert_to_datetime(date, tzinfo) for date in dates]

def compare_greater(date1, date2):
    """Compare two dates if date1 > date2 and make them comparable before."""
    date1, date2 = make_comparable((date1, date2))
    return date1 > date2

class Repetition:
    """A repetition of an event."""

    ATTRIBUTES_TO_DELETE_ON_COPY = [
        "RRULE", "RDATE", "EXDATE"
    ]

    def __init__(self, source, start, stop, keep_recurrence_attributes=False):
        self.source = source
        self.start = start
        self.stop = stop
        self.keep_recurrence_attributes = keep_recurrence_attributes

    def as_vevent(self):
        revent = self.source.copy()
        revent["DTSTART"] = vDDDTypes(self.start)
        revent["DTEND"] = vDDDTypes(self.stop)
        if not self.keep_recurrence_attributes:
            for attribute in self.ATTRIBUTES_TO_DELETE_ON_COPY:
                if attribute in revent:
                    del revent[attribute]
        for subcomponent in self.source.subcomponents:
            revent.add_component(subcomponent)
        return revent

    def is_in_span(self, span_start, span_stop):
        return time_span_contains_event(span_start, span_stop, self.start, self.stop)

    def __repr__(self):
        return "{}({{'UID':{}...}}, {}, {})".format(self.__class__.__name__, self.source.get("UID"), self.start, self.stop)

class RepeatedEvent:
    """An event with repetitions created from an ical event."""

    def __init__(self, event, keep_recurrence_attributes=False):
        self.event = event
        self.start = self.original_start = event["DTSTART"].dt
        self.end = self.original_end = self._get_event_end()
        self.keep_recurrence_attributes = keep_recurrence_attributes
        self.exdates = []
        self.exdates_utc = set()
        exdates = event.get("EXDATE", [])
        for exdates in ((exdates,) if not isinstance(exdates, list) else exdates):
            for exdate in exdates.dts:
                self.exdates.append(exdate.dt)
                self.exdates_utc.add(self._unify_exdate(exdate.dt))
        self.rdates = []
        rdates = event.get("RDATE", [])
        for rdates in ((rdates,) if not isinstance(rdates, list) else rdates):
            for rdate in rdates.dts:
                self.rdates.append(rdate.dt)

        self.make_all_dates_comparable()

        self.duration = self.end - self.start
        self.rule = rule = rruleset(cache=True)
        _rule = event.get("RRULE", None)
        if _rule:
            self.rrule = self.create_rule_with_start(_rule.to_ical().decode())
            rule.rrule(self.rrule)
        else:
            self.rrule = None

        for exdate in self.exdates:
            rule.exdate(exdate)
        for rdate in self.rdates:
            rule.rdate(rdate)
        rule.rdate(self.start)

    def create_rule_with_start(self, rule_string):
        try:
            return rrulestr(rule_string, dtstart=self.start)
        except ValueError:
            # string: FREQ=WEEKLY;UNTIL=20191023;BYDAY=TH;WKST=SU
            # start: 2019-08-01 14:00:00+01:00
            # ValueError: RRULE UNTIL values must be specified in UTC when DTSTART is timezone-aware
            rule_list = rule_string.split(";UNTIL=")
            assert len(rule_list) == 2
            date_end_index = rule_list[1].find(";")
            if date_end_index == -1:
                date_end_index = len(rule_list[1])
            until_string = rule_list[1][:date_end_index]
            if self.is_all_dates:
                until_string = until_string[:8]
            elif self.tzinfo is None:
                # remove the Z from the time zone
                until_string = until_string[:-1]
            else:
                # we assume the start is timezone aware but the until value is not, see the comment above
                if len(until_string) == 8:
                    until_string += "T000000"
                assert len(until_string) == 15
                until_string += "Z" # https://stackoverflow.com/a/49991809
            new_rule_string = rule_list[0] + rule_list[1][date_end_index:] + ";UNTIL=" + until_string
            return rrulestr(new_rule_string, dtstart=self.start)


    def make_all_dates_comparable(self):
        """Make sure we can use all dates with eachother.

        Dates may be mixed and we have many of them.
        - date
        - datetime without timezome
        - datetime with timezone
        These three are not comparable but can be converted.
        """
        self.tzinfo = None
        dates = [self.start, self.end] + self.exdates + self.rdates
        self.is_all_dates = not any(isinstance(date, datetime.datetime) for date in dates)
        for date in dates:
            if isinstance(date, datetime.datetime) and date.tzinfo is not None:
                self.tzinfo = date.tzinfo
                break
        self.start = convert_to_datetime(self.start, self.tzinfo)
        
        self.end = convert_to_datetime(self.end, self.tzinfo)
        self.rdates = [convert_to_datetime(rdate, self.tzinfo) for rdate in self.rdates]
        self.exdates = [convert_to_datetime(exdate, self.tzinfo) for exdate in self.exdates]

    def _unify_exdate(self, dt):
        """Return a unique string for the datetime which is used to
        compare it with exdates.
        """
        if isinstance(dt, datetime.datetime):
            return timestamp(dt)
        return dt

    def within_days(self, span_start, span_stop):
        # make dates comparable, rrule converts them to datetimes
        span_start = convert_to_datetime(span_start, self.tzinfo)
        span_stop = convert_to_datetime(span_stop, self.tzinfo)
        if compare_greater(span_start, self.start):
            # do not exclude an event if it spans across the time span
            span_start -= self.duration
        # NOTE: If in the following line, we get an error, datetime and date
        # may still be mixed because RDATE, EXDATE, start and rule.
        for start in self.rule.between(span_start, span_stop, inc=True):
            if isinstance(start, datetime.datetime) and start.tzinfo is not None:
                # update the time zone in case of summer/winter time change
                start = start.tzinfo.localize(start.replace(tzinfo=None))
            if self._unify_exdate(start) in self.exdates_utc:
                continue
            stop = start + self.duration
            yield Repetition(
                self.event,
                self.convert_to_original_type(start),
                self.convert_to_original_type(stop),
                self.keep_recurrence_attributes,
            )

    def convert_to_original_type(self, date):
        if not isinstance(self.original_start, datetime.datetime) and \
                not isinstance(self.original_end, datetime.datetime):
            return convert_to_date(date)
        return date

    def _get_event_end(self):
        end = self.event.get("DTEND")
        if end is not None:
            return end.dt
        duration = self.event.get("DURATION")
        if duration is not None:
            return self.event["DTSTART"].dt + duration.dt
        return self.event["DTSTART"].dt

    def is_recurrence(self):
        return "RECURRENCE-ID" in self.event

    def as_single_event(self):
        for repetition in self.within_days(self.start, self.start):
            return repetition


class UnfoldableCalendar:
    '''A calendar that can unfold its events at a certain time.'''

    def __init__(self, calendar, keep_recurrence_attributes=False):
        """Create an unfoldable calendar from a given calendar."""
        assert calendar.get("CALSCALE", "GREGORIAN") == "GREGORIAN", "Only Gregorian calendars are supported." # https://www.kanzaki.com/docs/ical/calscale.html
        self.repetitions = []
        for event in calendar.walk():
            if not is_event(event):
                continue
            self.repetitions.append(RepeatedEvent(event, keep_recurrence_attributes))

    @staticmethod
    def to_datetime(date):
        """Convert date inputs of various sorts into a datetime object."""
        if isinstance(date, tuple):
            return datetime.datetime(*date)
        elif isinstance(date, str):
            # see https://docs.python.org/2/library/datetime.html#strftime-strptime-behavior
            if len(date) == 8:
                return datetime.datetime.strptime(date, "%Y%m%d")
            return datetime.datetime.strptime(date, "%Y%m%dT%H%M%SZ")
        return date

    def all(self):
        """Returns all events."""
        # TODO: test MAX and MIN values
        return self.between((1000, 1, 1), (3000, 1, 1))

    _DELTAS = [
        datetime.timedelta(days=1),
        datetime.timedelta(hours=1),
        datetime.timedelta(minutes=1),
        datetime.timedelta(seconds=1)
    ]

    def at(self, date):
        """Return all events within the next 24 hours of starting at the given day.

        - date can be a year like (2019,) or 2099
        - a month like (2019, 1) for January of 2019
        - a day like (2019, 1, 1) for the first of January 2019
        - a day with hours, (2019, 1, 1, 1)
        - a day with minutes, (2019, 1, 1, 1, 1)
        - a day with seconds, (2019, 1, 1, 1, 1, 1)
        """
        if isinstance(date, int):
            date = (date,)
        if isinstance(date, str):
            assert len(date) == 8 and date.isdigit(), "format yyyymmdd expected"
            date = (int(date[:4], 10), int(date[4:6], 10), int(date[6:]))
        if isinstance(date, datetime.datetime):
            return self.between(date, date)
        if isinstance(date, datetime.date):
            return self.between(date, date + datetime.timedelta(days=1))
        if len(date) == 1:
            return self.between((date[0], 1, 1), (date[0] + 1, 1, 1))
        if len(date) == 2:
            year, month = date
            if month == 12:
                return self.between((year, 12, 1), (year + 1, 1, 1))
            return self.between((year, month, 1), (year, month + 1, 1))
        dt = self.to_datetime(date)
        return self.between(dt, dt + self._DELTAS[len(date) - 3])
    
    def between(self, start, stop): # TODO: add parameters from time_span_contains_event
        """Return events at a time between start (inclusive) and end (inclusive)"""
        span_start = self.to_datetime(start)
        span_stop = self.to_datetime(stop)
        events = []
        events_by_id = defaultdict(dict) # UID (str) : RECURRENCE-ID(datetime) : event (Event)
        default_uid = object()
        def add_event(event):
            """Add an event and check if it was edited."""
            same_events = events_by_id[event.get("UID", default_uid)] # TODO: test what comes first
            recurrence_id = event.get("RECURRENCE-ID", event["DTSTART"]).dt # TODO: this is still wrong: what if there are different events at the same time?
            if isinstance(recurrence_id, datetime.datetime):
                recurrence_id = recurrence_id.date()
            other = same_events.get(recurrence_id, None)
            if other: # TODO: test that this is independet of order
                event_recurrence_id = event.get("RECURRENCE-ID", None)
                other_recurrence_id = other.get("RECURRENCE-ID", None)
                if event_recurrence_id is not None and other_recurrence_id is None:
                    events.remove(other)
                elif event_recurrence_id is None and other_recurrence_id is not None:
                    return
                else:
                    event_sequence = event.get("SEQUENCE", None)
                    other_sequence = other.get("SEQUENCE", None)
                    if event_sequence is not None and other_sequence is not None:
                        if event["SEQUENCE"] < other["SEQUENCE"]:
                            return
                        events.remove(other)
            same_events[recurrence_id] = event
            events.append(event)

        span_start_day = span_start
        if isinstance(span_start_day, datetime.datetime):
            span_start_day = span_start_day.replace(hour=0,minute=0,second=0)
        span_stop_day = span_stop
        if isinstance(span_stop_day, datetime.datetime):
            span_stop_day = span_stop.replace(hour=23,minute=59,second=59)

        # repetitions must be considered because the may remove events from the time span
        # see https://github.com/niccokunzmann/python-recurring-ical-events/issues/62
        remove_because_not_in_span = []
        for event_repetitions in self.repetitions:
            if event_repetitions.is_recurrence():
                repetition = event_repetitions.as_single_event()
                vevent = repetition.as_vevent()
                add_event(vevent)
                if not repetition.is_in_span(span_start, span_stop):
                    remove_because_not_in_span.append(vevent)
                continue
            for repetition in event_repetitions.within_days(span_start_day, span_stop_day):
                if compare_greater(repetition.start, span_stop):
                    break
                if repetition.is_in_span(span_start, span_stop):
                    add_event(repetition.as_vevent())

        for vevent in remove_because_not_in_span:
            try:
                events.remove(vevent)
            except ValueError:
                pass

        return events

def of(a_calendar, keep_recurrence_attributes=False):
    """Unfold recurring events of a_calendar"""
    return UnfoldableCalendar(a_calendar, keep_recurrence_attributes)




import pytest
from datetime import date, datetime


@pytest.mark.parametrize("a_date,count", [
    # Year
    (2018, 3), ((2018,), 3),
    # Month
    ((2018, 1), 3),
    # Day
    ((2018, 1, 11), 1), (date(2018, 1, 11), 1), ("20180111", 1),
    ((2018, 1, 9), 0), (date(2018, 1, 9), 0), ("20180109", 0),
    # Datetime
    (datetime(2018, 1, 11, 10, 0, 0), 1),
    (datetime(2018, 1, 9, 10, 0, 0), 0)
])
def test_at_input_arguments(a_date, count, calendars):
    events = calendars.duration.at(a_date)
    assert len(events) == count



"""
This file contains the test cases which test that the
event uses the right class: date or datetime.
See https://github.com/niccokunzmann/python-recurring-ical-events/issues/7
"""
import pytest
import datetime
from icalendar.prop import vDatetime, vDate
from icalendar import Event

def test_can_serialize(calendars):
    """Test that the event can be serialized."""
    events = calendars.one_day_event.all()
    assert events
    event = events[0]
    string = event.to_ical()
    assert isinstance(string, bytes)

@pytest.mark.parametrize("attribute,dt_type,event_name", [
        ("dtstart", datetime.date,  "one_day_event"),
        ("dtend", datetime.date, "one_day_event"),
        ("dtstart", datetime.datetime, "one_event"),
        ("dtend", datetime.datetime, "one_event"),
    ])
@pytest.mark.parametrize("dt_or_content", [False])
def test_is_date(calendars, attribute, dt_type, event_name, dt_or_content):
    """Check the type of the attributes"""
    calendar = calendars[event_name]
    events = calendar.all()
    assert len(events) == 1
    event = events[0]
    event = Event.from_ical(event.to_ical())
    dt = event[attribute]
    assert isinstance(dt.dt, dt_type), "content of ical should match"




import pytest

def test_event_is_not_included_if_it_is_later(calendars):
    events = calendars.three_events.between((2000, 1, 1), (2001,1,1))
    assert not events

def test_event_is_not_included_if_it_is_earlier(calendars):
    events = calendars.three_events.between((2099, 1, 1), (2100,1,1))
    assert not events

def test_all_events_in_time_span(calendars):
    events = calendars.three_events.between((2000, 1, 1), (2100,1,1))
    assert len(events) == 3

@pytest.mark.parametrize("count,end",[
    (0, (2019, 3, 3)),
    (1, (2019, 3, 5)),
    (2, (2019, 3, 8)),
])
def test_events_occur_after_and_before_span_end(calendars, count, end):
    events = calendars.three_events.between((2000, 1, 1), end)
    assert len(events) == count

def test_include_events_if_the_time_zone_differs(todo):
    """When the time zone is different, events can be included or
    excluded because they are in another time zone.
    This may only be a problem for tests though because users will
    probably use time zones if it really makes a difference.
    """

@pytest.mark.parametrize("count,start",[
    (3, (2019, 3, 3)),
    (2, (2019, 3, 5)),
    (1, (2019, 3, 8)),
])
def test_events_occur_after_and_before_span_start(calendars, count, start):
    events = calendars.three_events.between(start, (2100, 1, 1))
    assert len(events) == count



'''
This tests the move of a december event.

Issue: https://github.com/niccokunzmann/python-recurring-ical-events/issues/62
'''

def test_event_is_absent(calendars):
    '''RRULE:FREQ=MONTHLY;BYDAY=-1FR'''
    events = calendars.issue_62_moved_event.at("20211231")
    assert events == []

def test_event_has_moved(calendars):
    '''DTSTART;TZID=Europe/Berlin:20211217T213000'''
    events = calendars.issue_62_moved_event.at("20211217")
    assert len(events) == 1

def test_there_is_only_one_event_in_december(calendars):
    '''Maybe, if we get the whole December, there might be one event.'''
    events = calendars.issue_62_moved_event.at((2021, 12))
    assert len(events) == 1

    





from recurring_ical_events import time_span_contains_event
import pytest
from datetime import datetime, date
from pytz import timezone, utc

berlin = timezone("Europe/Berlin")

@pytest.mark.parametrize(
    "span_start,span_stop,event_start,event_stop,"
    "include_start,include_stop,result,message", [

    (0, 5, 1, 3, True, True, True, "event lies inside 1"),
    (0, 5, 1, 3, False, True, True, "event lies inside 2"),
    (0, 5, 1, 3, True, False, True, "event lies inside 3"),
    (0, 5, 1, 3, False, False, True,"event lies inside 4"),

    (0, 5, -3, -1, True, True, False, "event lies before 1"),
    (0, 5, -3, -1, False, True, False, "event lies before 2"),
    (0, 5, -3, -1, True, False, False, "event lies before 3"),
    (0, 5, -3, -1, False, False, False, "event lies before 4"),

    (0, 5, 34, 41, True, True, False, "event lies after 1"),
    (0, 5, 34, 41, False, True, False, "event lies after 2"),
    (0, 5, 34, 41, True, False, False, "event lies after 3"),
    (0, 5, 34, 41, False, False, False, "event lies after 4"),

    (0, 5, 4, 6, True, True, True, "event overlaps end 1"),
    (0, 5, 4, 6, False, True, True, "event overlaps end 2"),
    (0, 5, 4, 6, True, False, False, "event overlaps end 3"),
    (0, 5, 4, 6, False, False, False, "event overlaps end 4"),

    (0, 5, -1, 2, True, True, True, "event overlaps start 1"),
    (0, 5, -1, 2, False, True, False, "event overlaps start 2"),
    (0, 5, -1, 2, True, False, True, "event overlaps start 3"),
    (0, 5, -1, 2, False, False, False, "event overlaps start 4"),

    (0, 5, -1, 7, True, True, True, "event overlaps start and end 1"),
    (0, 5, -1, 7, False, True, False, "event overlaps start and end 2"),
    (0, 5, -1, 7, True, False, False, "event overlaps start and end 3"),
    (0, 5, -1, 7, False, False, False, "event overlaps start and end 4"),

    (0, 5, 0, 4, True, True, True, "event begins at start 1"),
    (0, 5, 0, 4, False, True, True, "event begins at start 2"),
    (0, 5, 0, 4, True, False, True, "event begins at start 3"),
    (0, 5, 0, 4, False, False, True, "event begins at start 4"),

    (0, 5, 1, 5, True, True, True, "event ends at end 1"),
    (0, 5, 1, 5, False, True, True, "event ends at end 2"),
    (0, 5, 1, 5, True, False, True, "event ends at end 3"),
    (0, 5, 1, 5, False, False, True, "event ends at end 4"),

    (datetime(2019, 1, 1, 1), datetime(2019, 1, 4, 1), date(2019, 1, 2), date(2019, 1, 3), True, True, True, "date is in datetime span 1"),
    (datetime(2019, 1, 1, 1), datetime(2019, 1, 4, 1), date(2019, 1, 2), date(2019, 1, 3), False, True, True, "date is in datetime span 2"),
    (datetime(2019, 1, 1, 1), datetime(2019, 1, 4, 1), date(2019, 1, 2), date(2019, 1, 3), True, False, True, "date is in datetime span 3"),
    (datetime(2019, 1, 1, 1), datetime(2019, 1, 4, 1), date(2019, 1, 2), date(2019, 1, 3), False, False, True, "date is in datetime span 4"),

    (date(2019, 1, 3), date(2019, 1, 4), date(2019, 1, 2), date(2019, 1, 3), True, True, False, "date is in span 1"),
    (date(2019, 1, 3), date(2019, 1, 4), date(2019, 1, 2), date(2019, 1, 3), False, True, False, "date is in span 2"),
    (date(2019, 1, 3), date(2019, 1, 4), date(2019, 1, 2), date(2019, 1, 3), True, False, False, "date is in span 3"),
    (date(2019, 1, 3), date(2019, 1, 4), date(2019, 1, 2), date(2019, 1, 3), False, False, False, "date is in span 4"),

    (date(2019, 1, 4), date(2019, 1, 4), datetime(2019, 1, 2, 1), datetime(2019, 1, 4), True, True, False, "datetime is in date span start 1"),
    (date(2019, 1, 4), date(2019, 1, 4), datetime(2019, 1, 2, 1), datetime(2019, 1, 4), False, True, False, "datetime is in date span start 2"),
    (date(2019, 1, 4), date(2019, 1, 4), datetime(2019, 1, 2, 1), datetime(2019, 1, 4), True, False, False, "datetime is in date span start 3"),
    (date(2019, 1, 4), date(2019, 1, 4), datetime(2019, 1, 2, 1), datetime(2019, 1, 4), False, False, False, "datetime is in date span start 4"),

    (date(2019, 1, 4), date(2019, 1, 5), datetime(2019, 1, 4, 1), datetime(2019, 1, 4, 2), True, True, True, "datetime is in date span end 1"),
    (date(2019, 1, 4), date(2019, 1, 5), datetime(2019, 1, 4, 1), datetime(2019, 1, 4, 2), False, True, True, "datetime is in date span end 2"),
    (date(2019, 1, 4), date(2019, 1, 5), datetime(2019, 1, 4, 1), datetime(2019, 1, 4, 2), True, False, True, "datetime is in date span end 3"),
    (date(2019, 1, 4), date(2019, 1, 5), datetime(2019, 1, 4, 1), datetime(2019, 1, 4, 2), False, False, True, "datetime is in date span end 4"),

    (datetime(2019, 1, 1, 1, tzinfo=berlin), datetime(2019, 1, 4, 2, tzinfo=berlin), datetime(2019, 1, 4, 1, 10), datetime(2019, 1, 4, 1, 20), True, True, True, "without time zone is put into time zone 1"),
    (datetime(2019, 1, 1, 1, tzinfo=berlin), datetime(2019, 1, 4, 2, tzinfo=berlin), datetime(2019, 1, 4, 3, 10), datetime(2019, 1, 4, 3, 20), True, True, False, "without time zone is put into time zone 2"),
    (datetime(2019, 1, 1, 1), datetime(2019, 1, 4, 2), datetime(2019, 1, 4, 1, 10, tzinfo=berlin), datetime(2019, 1, 4, 1, 20, tzinfo=berlin), True, True, True, "without time zone is put into time zone 3"),
    (datetime(2019, 1, 1, 1), datetime(2019, 1, 4, 2), datetime(2019, 1, 4, 3, 10, tzinfo=berlin), datetime(2019, 1, 4, 3, 20, tzinfo=berlin), True, True, False, "without time zone is put into time zone 4"),

    (datetime(2019, 1, 1, 1, tzinfo=berlin), datetime(2019, 1, 1, 2, tzinfo=berlin), datetime(2019, 1, 1, 0, tzinfo=utc), datetime(2019, 1, 1, 1, tzinfo=utc), True, True, True, "comparing times from different time zones 1"),
    (datetime(2019, 1, 1, 4, tzinfo=berlin), datetime(2019, 1, 1, 5, tzinfo=berlin), datetime(2019, 1, 1, 0, tzinfo=utc), datetime(2019, 1, 1, 1, tzinfo=utc), True, True, False, "comparing times from different time zones 2"),
# The end of the VEVENT is exclusive, see RFC5545
#    Note that the "DTEND" property is
#    set to July 9th, 2007, since the "DTEND" property specifies the
#    non-inclusive end of the event.
#
#Tests:
#- We should not include an event which ends at a requested start date.
#- We should include an event which ends at an end date but is included
#  in the range.
    (date(2019, 1, 4), date(2019, 1, 5), date(2019, 1, 3), date(2019, 1, 4), True, True, False, "We should not include an event which ends at a requested start date."),
    (datetime(2019, 1, 4), datetime(2019, 1, 5), date(2019, 1, 3), date(2019, 1, 4), True, True, False, "We should not include an event which ends at a requested start date."),

    (datetime(2019, 1, 4), datetime(2019, 1, 5), datetime(2019, 1, 4, 3), datetime(2019, 1, 5), True, True, True, "We should include an event which ends at an end date but is included"),
    (date(2019, 1, 4), date(2019, 1, 4), date(2019, 1, 4), date(2019, 1, 4), True, True, True, "We should include an event which ends at an end date but is included"),
    (date(2019, 1, 3), date(2019, 1, 4), date(2019, 1, 4), date(2019, 1, 4), True, True, True, "We should include an event which ends at an end date but is included"),
])
def test_time_span_inclusion(
    span_start, span_stop, event_start, event_stop, include_start, include_stop,
    result, message):
    assert time_span_contains_event(
        span_start, span_stop, event_start, event_stop,
        include_start, include_stop
        ) == result, message

def test_event_of_zero_size(todo):
    """We should test events of zero size, ending and starting at the same time."""





def test_calendar_between_allows_tuple(todo):
    pass

def test_calendar_between_allows_float(todo):
    pass

def test_calendar_between_allows_date(todo):
    pass

def test_calendar_between_allows_datetime(todo):
    pass
